@@ -21,24 +21,25 @@
 
 package Lintian::debian::changelog;
 
-use strict;
+use v5.20;
 use warnings;
+use utf8;
 use autodie;
 
 use Date::Format qw(time2str);
 use Email::Valid;
-use Encode qw(decode);
 use List::Util qw(first);
 use List::MoreUtils qw(any uniq);
 use Path::Tiny;
 use Try::Tiny;
+use Unicode::UTF8 qw(valid_utf8 decode_utf8);
 
 use Lintian::Data ();
 use Lintian::Inspect::Changelog;
 use Lintian::Inspect::Changelog::Version;
 use Lintian::Relation::Version qw(versions_gt);
 use Lintian::Spelling qw(check_spelling);
-use Lintian::Util qw(file_is_encoded_in_non_utf8 strip);
+use Lintian::Util qw(strip);
 
 use constant EMPTY => q{};
 
@@ -66,8 +67,11 @@ sub source {
     my $processable = $self->processable;
     my $group = $self->group;
 
-    my @entries = @{$processable->changelog->entries};
+    my $changelog = $processable->changelog;
+    return
+      unless defined $changelog;
 
+    my @entries = @{$changelog->entries};
     return
       unless @entries;
 
@@ -355,36 +359,41 @@ sub binary {
     my $news;
     my $dnews = path($processable->groupdir)->child('NEWS.Debian')->stringify;
     if (-f $dnews) {
-        my $line = file_is_encoded_in_non_utf8($dnews);
-        if ($line) {
-            $self->tag('debian-news-file-uses-obsolete-national-encoding',
-                "at line $line");
-        }
-        my $changelog = Lintian::Inspect::Changelog->new;
-        my $contents = path($dnews)->slurp;
-        $changelog->parse($contents);
 
-        if (my @errors = @{$changelog->errors}) {
-            for (@errors) {
-                $self->tag('syntax-error-in-debian-news-file',
-                    "line $_->[0]","\"$_->[1]\"");
-            }
-        }
+        my $bytes = path($dnews)->slurp;
+        if (valid_utf8($bytes)) {
 
-        # Some checks on the most recent entry.
-        if ($changelog->entries && defined @{$changelog->entries}[0]) {
-            ($news) = @{$changelog->entries};
-            if ($news->Distribution && $news->Distribution eq 'UNRELEASED') {
-                $self->tag('debian-news-entry-has-strange-distribution',
-                    $news->Distribution);
+            my $contents = decode_utf8($bytes);
+            my $changelog = Lintian::Inspect::Changelog->new;
+            $changelog->parse($contents);
+
+            if (my @errors = @{$changelog->errors}) {
+                for (@errors) {
+                    $self->tag('syntax-error-in-debian-news-file',
+                        "line $_->[0]","\"$_->[1]\"");
+                }
             }
-            check_spelling(
-                $news->Changes,
-                $group->spelling_exceptions,
-                $self->spelling_tag_emitter('spelling-error-in-news-debian'));
-            if ($news->Changes =~ /^\s*\*\s/) {
-                $self->tag('debian-news-entry-uses-asterisk');
+
+            # Some checks on the most recent entry.
+            if ($changelog->entries && defined @{$changelog->entries}[0]) {
+                ($news) = @{$changelog->entries};
+                if ($news->Distribution && $news->Distribution eq 'UNRELEASED')
+                {
+                    $self->tag('debian-news-entry-has-strange-distribution',
+                        $news->Distribution);
+                }
+                check_spelling(
+                    $news->Changes,
+                    $group->spelling_exceptions,
+                    $self->spelling_tag_emitter(
+                        'spelling-error-in-news-debian'));
+                if ($news->Changes =~ /^\s*\*\s/) {
+                    $self->tag('debian-news-entry-uses-asterisk');
+                }
             }
+
+        } else {
+            $self->tag('debian-news-file-uses-obsolete-national-encoding');
         }
     }
 
@@ -490,10 +499,10 @@ sub binary {
     }
 
     # check that changelog is UTF-8 encoded
-    my $line = file_is_encoded_in_non_utf8($dchpath);
-    if ($line) {
-        $self->tag('debian-changelog-file-uses-obsolete-national-encoding',
-            "at line $line");
+    my $bytes = path($dchpath)->slurp;
+    unless (valid_utf8($bytes)) {
+        $self->tag('debian-changelog-file-uses-obsolete-national-encoding');
+        return;
     }
 
     my $changelog = $processable->changelog;
@@ -680,16 +689,10 @@ sub binary {
             }
         }
 
-        # We have to decode into UTF-8 to get the right length for the
-        # length check.  For some reason, use open ':utf8' isn't
-        # sufficient.  If the changelog uses a non-UTF-8 encoding,
-        # this will mangle it, but it doesn't matter for the length
-        # check.
-        #
         # Parse::DebianChangelog adds an additional space to the
         # beginning of each line, so we have to adjust for that in the
         # length check.
-        my @lines = split("\n", decode('utf-8', $changes));
+        my @lines = split(/\n/, $changes);
         for my $i (0 .. $#lines) {
             my $line = $i + $chloff;
             $self->tag('debian-changelog-line-too-short', $1, "(line $line)")
@@ -727,7 +730,7 @@ sub check_dch {
     my ($prefix, $suffix);
     my $lineno = 0;
     my ($estart, $tstart) = (0, 0);
-    open(my $fd, '<', $path);
+    open(my $fd, '<:encoding(UTF-8)', $path);
     while (<$fd>) {
 
         unless ($tstart) {
